import functools
import operator
import re
import uuid
import logging
from copy import deepcopy
from collections import defaultdict
from itertools import zip_longest
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Set, Tuple, Union

from picblocks.blockhasher import BlockHasher
from mcrit.index.SearchCursor import FullSearchCursor

from mcrit.index.SearchQueryTree import AndNode, BaseVisitor, FilterSingleElementLists, NodeType, OrNode, PropagateNot, SearchConditionNode, SearchFieldResolver
from mcrit.minhash.MinHash import MinHash
from mcrit.storage.FamilyEntry import FamilyEntry
from mcrit.storage.FunctionEntry import FunctionEntry
from mcrit.storage.FunctionLabelEntry import FunctionLabelEntry
from mcrit.storage.MatchingCache import MatchingCache
from mcrit.storage.SampleEntry import SampleEntry
from mcrit.storage.StorageInterface import StorageInterface

if TYPE_CHECKING: # pragma: no cover
    from mcrit.config.McritConfig import McritConfig
    from mcrit.config.StorageConfig import StorageConfig
    from mcrit.config.MinHashConfig import MinHashConfig
    from smda.common.SmdaFunction import SmdaFunction
    from smda.common.SmdaReport import SmdaReport

LOGGER = logging.getLogger(__name__)


# Helper Methods for search
def _get_field_once(object, field):
    try:
        return getattr(object, field)
    except Exception:
        pass
    try:
        return object[field]
    except Exception:
        pass

def _get_field(object, field):
    dot_index = field.find(".")
    if dot_index >= 0:
        first_field = field[:dot_index]
        remaining_fields = field[dot_index+1:]
        return _get_field(
            _get_field_once(object, first_field),
            remaining_fields
        )
    else:
        return _get_field_once(object, field)


class MemorySearchTranspiler(BaseVisitor):
    """
    Converts a tree to a filter function that can be applied to an entry, returning True or False.
    The input tree MUST NOT contain Not or SearchTerm nodes.
    """
    def visitAndNode(self, node:AndNode) -> Callable:
        visited_children: List[Callable] = [self.visit(child) for child in node.children]
        def and_function(entry):
            for function in visited_children:
                if not function(entry):
                    return False
            return True
        return and_function

    def visitOrNode(self, node:OrNode) -> Callable:
        visited_children: List[Callable] = [self.visit(child) for child in node.children]
        def or_function(entry):
            for function in visited_children:
                if function(entry):
                    return True
            return False
        return or_function

    def visitSearchConditionNode(self, node:SearchConditionNode) -> Callable:
        string_to_operator = {
            "<": operator.lt,
            "<=": operator.le,
            ">": operator.gt,
            ">=": operator.ge,
            "=": operator.eq,
            "": operator.eq,
            "!=": operator.ne,
            "?": None,
            "!?": None,
        }
        value = node.value
        if node.operator.endswith("?"):
            regex = re.compile(re.escape(node.value), re.IGNORECASE)
            inverse = node.operator == "!?"
            def check_regex(entry):
                value = _get_field(entry, node.field)
                if not isinstance(value, str):
                    return False
                return (regex.search(value) is not None) ^ inverse
            return check_regex

        if node.field in ("pichash", "offset") or node.field.endswith("_id") or "num_" in node.field:
            try:
                value = int(value, 0)
            except Exception:
                pass

        chosen_operator = string_to_operator[node.operator]
        def compare(entry):
            return chosen_operator(_get_field(entry, node.field), value)
        return compare

class MemoryStorage(StorageInterface):
    _families: Dict[int, FamilyEntry]
    _samples: Dict[int, "SampleEntry"]
    _functions: Dict[int, "FunctionEntry"]
    _query_samples: Dict[int, "SampleEntry"]
    _query_functions: Dict[int, "FunctionEntry"]
    _pichashes: Dict[int, Set[Tuple[int, int, int]]]
    # Dict[band_number, Dict[?, List[function_id]]]
    _bands: Dict[int, Dict[int, List[int]]]
    _counters: Dict[str, int]
    _sample_by_sha256: Dict[str, int]
    _sample_id_to_function_ids: Dict[int, List[int]]

    def __init__(self, config: "McritConfig") -> None:
        super().__init__(config)  # sets config
        self._setupEmptyStorage()
        self.blockhasher = BlockHasher()

    def clearStorage(self) -> None:
        self._setupEmptyStorage()

    def _setupEmptyStorage(self) -> None:
        self._mcrit_db_id = str(uuid.uuid4())
        self._db_state = 0
        self._families = {}
        self._samples = {}
        self._functions = {}
        self._query_samples = {}
        self._query_functions = {}
        self._pichashes = {}
        self._bands = {band_number: {} for band_number in range(self._storage_config.STORAGE_NUM_BANDS)}
        self._counters = defaultdict(lambda: 0)
        # initialize query sample/function ids 
        if self._counters["query_samples"] == 0:
            self._counters["query_samples"] += 1
        if self._counters["query_functions"] == 0:
            self._counters["query_functions"] += 1
        # caches
        self._sample_by_sha256 = {}
        self._sample_id_to_function_ids = defaultdict(list)
        unknown_family_id = self.addFamily("")
        assert unknown_family_id == 0

    def _updateDbState(self):
        self._db_state += 1

    def _useCounter(self, name: str) -> int:
        result = self._counters[name]
        self._counters[name] += 1
        return result

    def _updateFamilyStats(self, family_id, num_samples_inc, num_functions_inc, num_library_samples_inc):
        family_entry = self.getFamily(family_id)
        assert family_entry is not None
        family_entry.num_samples += num_samples_inc
        family_entry.num_functions += num_functions_inc
        family_entry.num_library_samples += num_library_samples_inc

    def dbLogEvent(self, event_msg, username=None, details=None):
        """ For MemoryStorage, this is a dummy function and we do not persist logging messages """
        logging.info(event_msg)

    # TODO check if this works
    def deleteSample(self, sample_id: int) -> bool:
        if not self.isSampleId(sample_id):
            return False
        function_ids = self._sample_id_to_function_ids[sample_id]
        if sample_id < 0:
            for function_id in function_ids:
                del self._functions[function_id]
            del self._sample_id_to_function_ids[sample_id]
            del self._samples[sample_id]
            return True
        for function_id in function_ids:
            function_entry = self._functions[function_id]
            minhash = function_entry.getMinHash(self._minhash_config.MINHASH_SIGNATURE_BITS)
            # remove function
            del self._functions[function_id]
            # remove pichash entries
            if function_entry.pichash:
                self._pichashes[function_entry.pichash].remove((function_entry.family_id, sample_id, function_id))
            # remove minhash entries, if necessary
            if not minhash or not minhash.hasMinHash():
                continue
            band_hashes = self.getBandHashesForMinHash(minhash)
            for band_number, band_hash in sorted(band_hashes.items()):
                if band_hash not in self._bands[band_number]:
                    continue
                # delete function id from bandhash
                self._bands[band_number][band_hash].remove(function_id)
                # delete bandhash if empty
                if not self._bands[band_number][band_hash]:
                    del self._bands[band_number][band_hash]
        # remove references
        del self._sample_id_to_function_ids[sample_id]
        # in case to samples with same sha256 are added and then removed, it could crash, so remove entry safely
        self._sample_by_sha256.pop(self._samples[sample_id].sha256, None)

        sample_entry = self.getSampleById(sample_id)
        self._updateFamilyStats(sample_entry.family_id, -1, -sample_entry.statistics["num_functions"], -int(sample_entry.is_library))
        # remove sample
        del self._samples[sample_id]
        return True

    def modifySample(self, sample_id: int, update_information: dict) -> bool:
        if not self.isSampleId(sample_id):
            return False
        if "is_library" in update_information:
            is_library_info_changed = self._samples[sample_id].is_library != update_information["is_library"]
            self._samples[sample_id].is_library = update_information["is_library"]
            if is_library_info_changed:
                self._families[family_id].num_library_samples += 1 if update_information["is_library"] else -1
        if "family_name" in update_information:
            family_name = update_information["family_name"]
            old_family_id = self._samples[sample_id].family_id
            family_id = self.addFamily(family_name)
            # update sample_entry and function_entries with new family information
            self._samples[sample_id].family = family_name
            self._samples[sample_id].family_id = family_id
            for function_id, function_entry in self._functions.items():
                if function_entry.sample_id == sample_id:
                    function_entry.family_id = family_id
                    self._functions[function_id] = function_entry  
                    self._pichashes[function_entry.pichash].remove((old_family_id, sample_id, function_id))
                    self._pichashes[function_entry.pichash].add((family_id, sample_id, function_id))
            # update family information itself
            self._families[old_family_id].num_samples -= 1
            self._families[old_family_id].num_functions -= self._samples[sample_id].statistics["num_functions"]
            self._families[family_id].num_samples += 1
            self._families[family_id].num_functions += self._samples[sample_id].statistics["num_functions"]
            if self._samples[sample_id].is_library:
                self._families[old_family_id].num_library_samples -= 1
                self._families[family_id].num_library_samples += 1
            if self._families[old_family_id].num_samples == 0 and old_family_id != 0:
                self._families.pop(old_family_id)
        if "version" in update_information:
            self._samples[sample_id].version = update_information["version"]
        if "component" in update_information:
            self._samples[sample_id].component = update_information["component"]
        return True

    def modifyFamily(self, family_id: int, update_information: dict) -> bool:
        if not self.isFamilyId(family_id):
            return False
        old_family_info = self.getFamily(family_id)
        if "is_library" in update_information:
            for sample_id, sample_entry in self._samples.items():
                if family_id == sample_entry.family_id:
                    self._samples["sample_id"].is_library = update_information["is_library"]
            self._families[family_id].num_library_samples = self._families[family_id].num_samples
        if "family_name" in update_information:
            old_family_info = self.getFamily(family_id)
            family_name = update_information["family_name"]
            new_family_id = self.addFamily(family_name)
            new_family_info = self.getFamily(new_family_id)
            new_num_samples = new_family_info.num_samples + old_family_info.num_samples
            new_num_functions = new_family_info.num_functions + old_family_info.num_functions
            new_num_lib_samples = new_family_info.num_library_samples + old_family_info.num_library_samples
            # update family_entry
            if family_id == 0:
                self._families[0].num_samples = 0
                self._families[0].num_functions = 0
                self._families[0].num_library_samples = 0
            else:
                self._families.pop(family_id)
            self._families[new_family_id].num_samples = new_num_samples
            self._families[new_family_id].num_functions = new_num_functions
            self._families[new_family_id].num_library_samples = new_num_lib_samples
            # update sample_entry and function_entries with new family information
            for sample_id, sample_entry in self._samples.items():
                if family_id == sample_entry.family_id:
                    self._samples[sample_id].family_id = new_family_id
                    self._samples[sample_id].family = family_name
            for function_id, function_entry in self._functions.items():
                if family_id == function_entry.family_id:
                    self._functions[function_id].family_id = new_family_id
                    self._pichashes[function_entry.pichash].remove((family_id, sample_id, function_id))
                    self._pichashes[function_entry.pichash].add((new_family_id, sample_id, function_id))
        self._updateDbState()
        return True

    def deleteFamily(self, family_id: int, keep_samples: Optional[str] = False) -> bool:
        if family_id not in self._families:
            return False
        sample_entries = self.getSamplesByFamilyId(family_id)
        if family_id == 0:
            self._families[0].num_samples = 0
            self._families[0].num_functions = 0
            self._families[0].num_library_samples = 0
        else:
            self._families.pop(family_id)
        if keep_samples:
            for sample_entry in sample_entries:
                self._samples[sample_entry.sample_id]["family_id"] = 0
                self._samples[sample_entry.sample_id]["family"] = ""
            function_ids_to_modify = set()
            for function_id, function_entry in self._functions.items():
                if function_entry.family_id == family_id:
                    function_ids_to_modify.add(function_id)
            for function_id in function_ids_to_modify:
                    self._functions[function_id]["family_id"] = 0
            self._families[0].num_samples += len(sample_entries)
            self._families[0].num_functions += len(function_ids_to_modify)
            self._families[0].num_library_samples += len([s for s in sample_entries if s.is_library])
        else:
            for sample_entry in sample_entries:
                self.deleteSample(sample_entry.sample_id)
        self._updateDbState()
        return True

    def updateFunctionLabels(self, smda_report: "SmdaReport", username: str) -> Optional["SampleEntry"]:
        sample_entry = self.getSampleBySha256(smda_report.sha256)
        if not sample_entry:
            return False
        # check which functions in the SmdaReport have suitable function_names
        extracted_labels = {}
        for smda_function in smda_report.getFunctions():
            function_name = smda_function.function_name
            if function_name and not re.match("sub_[a-fA-F0-9]{1,16}", function_name):
                extracted_labels[smda_function.offset] = function_name
        # get the respective FunctionEntries and check if the label is novel
        sample_function_entries = {entry.offset: entry for entry in self.getFunctionsBySampleId(sample_entry.sample_id)}
        label_updates = []
        for label_offset, extracted_label in extracted_labels.items():
            is_new_label = False
            if label_offset in sample_function_entries:
                is_new_label = True
                existing_labels = sample_function_entries[label_offset].function_labels
                for existing_label in existing_labels:
                    if existing_label.username == username and existing_label.function_label == extracted_label:
                        is_new_label = False
            # match by function_id or offset and add the label if it had not existed before.
            if is_new_label:
                new_function_entry_label = FunctionLabelEntry(extracted_label, username)
                self._functions[sample_function_entries[label_offset].function_id].function_labels.append(new_function_entry_label)

    def addSmdaReport(self, smda_report: "SmdaReport", isQuery=False) -> Optional["SampleEntry"]:
        sample_entry = None
        if isQuery:
            sample_entry = SampleEntry(
                smda_report, sample_id=-1 * self._useCounter("query_samples"), family_id=0
            )
            self._query_samples[sample_entry.sample_id] = sample_entry
            function_ids = []
            for smda_function in smda_report.getFunctions():
                function_entry = self._addFunction(sample_entry, smda_function, isQuery=True)
                function_ids.append(function_entry.function_id)
            self._sample_id_to_function_ids[sample_entry.sample_id] = function_ids
        else:
            if not self.getSampleBySha256(smda_report.sha256):
                family_id = self.addFamily(smda_report.family)
                sample_entry = SampleEntry(
                    smda_report, sample_id=self._useCounter("samples"), family_id=family_id
                )
                self._samples[sample_entry.sample_id] = sample_entry
                self._sample_by_sha256[sample_entry.sha256] = sample_entry.sample_id
                function_ids = []
                for smda_function in smda_report.getFunctions():
                    function_entry = self._addFunction(sample_entry, smda_function)
                    function_ids.append(function_entry.function_id)
                self._sample_id_to_function_ids[sample_entry.sample_id] = function_ids
                self._updateFamilyStats(family_id, +1, sample_entry.statistics["num_functions"], int(sample_entry.is_library))
            else:
                LOGGER.warning("Sample %s already existed, skipping.", smda_report.sha256)
        return sample_entry

    def importSampleEntry(self, sample_entry: "SampleEntry") -> Optional["SampleEntry"]:
        if not self.getSampleBySha256(sample_entry.sha256):
            sample_id = self._useCounter("samples")
            sample_entry.sample_id = sample_id
            self._samples[sample_id] = sample_entry
            self._sample_by_sha256[sample_entry.sha256] = sample_id
            self._updateFamilyStats(sample_entry.family_id, +1, sample_entry.statistics["num_functions"], int(sample_entry.is_library))
        else:
            LOGGER.warning("Sample %s already existed, skipping.", sample_entry.sha256)
        return sample_entry

    def importFunctionEntry(self, function_entry: "FunctionEntry") -> Optional["FunctionEntry"]:
        function_entry.function_id = self._useCounter("functions")
        if function_entry.function_name and len(function_entry.function_labels) == 0:
            new_function_entry_label = FunctionLabelEntry(function_entry.function_name, "mcrit-import")
            function_entry.function_labels.append(new_function_entry_label)
        # add function to regular storage
        self._functions[function_entry.function_id] = function_entry
        if function_entry.sample_id not in self._sample_id_to_function_ids:
            self._sample_id_to_function_ids[function_entry.sample_id] = []
        self._sample_id_to_function_ids[function_entry.sample_id].append(function_entry.function_id)
        if function_entry.pichash:
            if function_entry.pichash not in self._pichashes:
                self._pichashes[function_entry.pichash] = set()
            self._pichashes[function_entry.pichash].add((function_entry.family_id, function_entry.sample_id, function_entry.function_id))
        return function_entry

    def importFunctionEntries(self, function_entries: List["FunctionEntry"]) -> Optional[List["FunctionEntry"]]:
        for function_entry in function_entries:
            function_entry.function_id = self._useCounter("functions")
            if function_entry.function_name and len(function_entry.function_labels) == 0:
                new_function_entry_label = FunctionLabelEntry(function_entry.function_name, "mcrit-import")
                function_entry.function_labels.append(new_function_entry_label)
            # add function to regular storage
            self._functions[function_entry.function_id] = function_entry
            if function_entry.sample_id not in self._sample_id_to_function_ids:
                self._sample_id_to_function_ids[function_entry.sample_id] = []
            self._sample_id_to_function_ids[function_entry.sample_id].append(function_entry.function_id)
            if function_entry.pichash:
                if function_entry.pichash not in self._pichashes:
                    self._pichashes[function_entry.pichash] = set()
                self._pichashes[function_entry.pichash].add((function_entry.family_id, function_entry.sample_id, function_entry.function_id))
        return function_entries

    # TODO return type?
    def getLibraryInfoForSampleId(self, sample_id: int) -> Optional[Dict[str, str]]:
        if sample_id not in self._samples:
            return None
        sample_entry = self._samples[sample_id]
        return {"family": sample_entry.family, "version": sample_entry.version} if sample_entry.is_library else None

    def _addFunction(
        self, sample_entry: "SampleEntry", smda_function: "SmdaFunction", minhash: Optional["MinHash"] = None, isQuery=False
    ) -> "FunctionEntry":
        """Add a function (and optionally its MinHash) to storage, using the respective SampleEntry for reference.

        Args:
            sample_entry: The SampleEntry of smda_function
            smda_function: The SmdaFunction to be added.
            minhash: (Optional, defaults to None) The MinHash of the Function

        Returns:
            A FunctionEntry
        """
        if isQuery:
            function_entry = FunctionEntry(sample_entry, smda_function, -1 * self._useCounter("query_functions"), minhash=minhash)
            self._query_functions[function_entry.function_id] = function_entry
            return function_entry
        else:
            function_entry = FunctionEntry(sample_entry, smda_function, self._useCounter("functions"), minhash=minhash)
        image_lower = sample_entry.base_addr
        image_upper = image_lower + sample_entry.binary_size
        picblockhashes = []
        for hash_entry in self.blockhasher.getBlockhashesForFunction(smda_function, image_lower, image_upper, hash_size=8):
            for block_entry in hash_entry["offset_tuples"]:
                block_entry["hash"] = hash_entry["hash"]
                picblockhashes.append(block_entry)
        function_entry.picblockhashes = picblockhashes
        self._functions[function_entry.function_id] = function_entry
        if minhash and minhash.hasMinHash():
            minhash.function_id = function_entry.function_id
            self._addMinHashToBands(minhash)
        if function_entry.pichash:
            if function_entry.pichash not in self._pichashes:
                self._pichashes[function_entry.pichash] = set()
            self._pichashes[function_entry.pichash].add((function_entry.family_id, sample_entry.sample_id, function_entry.function_id))
        return function_entry

    def addMinHash(self, minhash: "MinHash") -> bool:
        if minhash.function_id is None or minhash.function_id not in self._functions:
            return False
        self._functions[minhash.function_id].minhash = minhash.getMinHash()
        self._functions[minhash.function_id].minhash_shingle_composition = minhash.getComposition()
        self._addMinHashToBands(minhash)
        return True

    def addMinHashes(self, minhashes: List["MinHash"]) -> None:
        for minhash in minhashes:
            self.addMinHash(minhash)

    def createMatchingCache(self, function_ids: List[int]) -> MatchingCache:
        # TODO: we might want add a flag to allow/disallow returning self
        cache_data = self._getCacheDataForFunctionIds(function_ids)
        return MatchingCache(cache_data)

    def _getCacheDataForFunctionIds(self, function_ids: List[int]) -> Dict:
        cache_data = {}
        sample_ids = {}
        sample_to_func_ids = {}
        minhashes = {}
        for function_id in function_ids:
            function_entry = self._functions[function_id]
            function_id = function_entry.function_id
            sample_id = function_entry.sample_id
            minhashes[function_id] = function_entry.minhash
            sample_ids[function_id] = sample_id
            if sample_id not in sample_to_func_ids:
                sample_to_func_ids[sample_id] = set()
            sample_to_func_ids[sample_id].add(function_id)
        cache_data["func_id_to_minhash"] = minhashes
        cache_data["func_id_to_sample_id"] = sample_ids
        cache_data["sample_id_to_func_ids"] = sample_to_func_ids
        return cache_data

    def clearMatchingCache(self) -> None:
        """no dedicated cache - no cleanup"""
        return

    def getSampleIds(self) -> List[int]:
        # TODO is deepcopy necessary?
        return deepcopy(list(self._samples.keys()))

    def getSamplesByFamilyId(self, family_id: int) -> Optional[List[SampleEntry]]:
        if family_id not in self._families:
            return None
        return [
            self._samples[sample_id] for sample_id in self._samples if self._samples[sample_id].family_id == family_id
        ]

    def getFamilyIds(self) -> List[int]:
        # TODO is deepcopy necessary?
        return deepcopy(list(self._families.keys()))

    def isSampleId(self, sample_id: int) -> bool:
        return sample_id in self._samples or sample_id in self._query_samples

    def isFunctionId(self, function_id: int) -> bool:
        return function_id in self._functions or function_id in self._query_functions

    def isFamilyId(self, family_id: int) -> bool:
        return family_id in self._families

    def isPicHash(self, pichash: int) -> bool:
        return pichash in self._pichashes

    def getMatchesForPicHash(self, pichash: int) -> Set[Tuple[int, int, int]]:
        if pichash not in self._pichashes:
            return set()
        return deepcopy(self._pichashes[pichash])

    def getMatchesForPicBlockHash(self, picblockhash: int) -> Set[Tuple[int, int, int, int]]:
        result = set()
        for function_id, function_entry in self._functions.items():
            for pbh in function_entry.picblockhashes:
                if picblockhash == pbh["hash"]:
                    result.add([function_entry.family_id, function_entry.sample_id, function_id, pbh["offset"]])
        return result

    def getFamily(self, family_id: int) -> Optional[FamilyEntry]:
        if family_id in self._families:
            return self._families[family_id]
        return None

    def getFamilyId(self, family_name: str) -> Optional[int]:
        for fam_id, fam_entry in self._families.items():
            if fam_entry.family_name == family_name:
                return fam_id
        return None

    def addFamily(self, family_name: str) -> int:
        family_id = self.getFamilyId(family_name)
        if family_id is None:
            family_id = self._useCounter("families")
            self._families[family_id] = FamilyEntry(family_name = family_name, family_id=family_id)
        return family_id

    def getFunctionById(self, function_id: int, with_xcfg=False) -> Optional["FunctionEntry"]:
        function_entry = None
        if function_id < 0:
            if function_id in self._query_functions:
                function_entry = deepcopy(self._query_functions[function_id])
        else:
            if function_id in self._functions:
                function_entry = deepcopy(self._functions[function_id])
        if function_entry:
            if with_xcfg is False:
                function_entry.xcfg = None
            return deepcopy(function_entry)
        return None

    # TODO does this need to be more efficient?
    def getFunctionsBySampleId(self, sample_id: int) -> Optional[List["FunctionEntry"]]:
        if sample_id in self._samples or sample_id in self._query_samples:
            function_ids = self._sample_id_to_function_ids[sample_id]
            function_entries = []
            for function_id in function_ids:
                if sample_id < 0:
                    function_entry = self._query_functions[function_id]
                else:
                    function_entry = self._functions[function_id]
                function_entries.append(function_entry)
            # TODO is deepcopy necessary?
            return deepcopy(function_entries)
        return None

    def getFunctionIdsBySampleId(self, sample_id: int) -> Optional[List["int"]]:
        function_ids = None
        if sample_id in self._samples or sample_id in self._query_samples:
            function_ids = self._sample_id_to_function_ids[sample_id]
        return function_ids

    def getFunctions(self, start_index: int, limit: int) -> Optional["FunctionEntry"]:
        index = 0
        function_entries = []
        for _, function_entry in self._functions.items():
            if index >= start_index:
                if (limit == 0) or (len(function_entries) < limit):
                    function_entries.append(function_entry)
                else:
                    break
            index += 1
        return deepcopy(function_entries)

    def getSampleById(self, sample_id: int) -> Optional["SampleEntry"]:
        if sample_id in self._samples:
            return deepcopy(self._samples[sample_id])
        if sample_id in self._query_samples:
            return deepcopy(self._query_samples[sample_id])

    def getSampleBySha256(self, sha256: str) -> Optional["SampleEntry"]:
        sample_entry = None
        if sha256 in self._sample_by_sha256:
            sample_id = self._sample_by_sha256[sha256]
            sample_entry = deepcopy(self._samples[sample_id])
        return sample_entry

    def getSampleIdByFunctionId(self, function_id: int) -> Optional[int]:
        sample_id = None
        if function_id in self._functions:
            sample_id = self._functions[function_id].sample_id
        if function_id in self._query_functions:
            sample_id = self._query_functions[function_id].sample_id
        return sample_id

    def getSamples(self, start_index: int, limit: int) -> Optional["SampleEntry"]:
        index = 0
        sample_entries = []
        for _, sample_entry in self._samples.items():
            if index >= start_index:
                if (limit == 0) or (len(sample_entries) < limit):
                    sample_entries.append(sample_entry)
                else:
                    break
            index += 1
        return deepcopy(sample_entries)

    def getPicHashMatchesByFunctionId(self, function_id: int) -> Optional[Dict[int, Set[Tuple[int, int]]]]:
        if function_id not in self._functions:
            return None
        pichash = self._functions[function_id].pichash
        if pichash is None:
            return None
        return {pichash: deepcopy(self._pichashes[pichash])}

    def getPicHashMatchesByFunctionIds(self, function_ids: List[int]) -> Dict[int, Set[Tuple[int, int]]]:
        pichashes = {}
        for function_id in function_ids:
            if function_id in self._functions:
                pichash = self._functions[function_id].pichash
                if pichash is None:
                    continue
                pichashes[pichash] = deepcopy(self._pichashes[pichash])
        return pichashes

    def getPicHashMatchesBySampleId(self, sample_id: int) -> Optional[Dict[int, Set[Tuple[int, int]]]]:
        function_entries = self.getFunctionsBySampleId(sample_id)
        if function_entries is None:
            return None
        function_ids = [fe.function_id for fe in function_entries]
        return self.getPicHashMatchesByFunctionIds(function_ids)

    def getMinHashByFunctionId(self, function_id: int) -> Optional[bytes]:
        if function_id in self._functions:
            function_entry = self._functions[function_id]
            return function_entry.minhash
        return None

    # -> Dict[function_id, Set[function_id]]
    # TODO optimize or move to interface
    def getCandidatesForMinHashes(self, function_id_to_minhash: Dict[int, "MinHash"], band_matches_required=1) -> Dict[int, Set[int]]:
        candidates = {}
        for function_id, minhash in function_id_to_minhash.items():
            candidates[function_id] = self.getCandidatesForMinHash(minhash, band_matches_required=band_matches_required)
        return candidates

    # -> Set[function_id]
    def getCandidatesForMinHash(self, minhash: "MinHash", band_matches_required=1) -> Set[int]:
        candidates = {}
        band_hashes = self.getBandHashesForMinHash(minhash)
        for band_number, band_hash in sorted(band_hashes.items()):
            if band_hash in self._bands[band_number]:
                for function_id in self._bands[band_number][band_hash]:
                    if function_id not in candidates:
                        candidates[function_id] = 0
                    candidates[function_id] += 1
        # reduce candidates based on banding requirements
        valid_candidates = set([])
        for function_id, hit_count in candidates.items():
            if hit_count >= band_matches_required:
                valid_candidates.add(function_id)
        return valid_candidates

    def getUnhashedFunctions(self, function_ids: Optional[List[int]] = None, only_function_ids=False) -> List["FunctionEntry"]:
        result = []
        if function_ids is None:
            if only_function_ids:
                result = [function_entry.function_id for _, function_entry in self._functions.items() if function_entry.xcfg and not function_entry.minhash]
            else:
                result = [function_entry for _, function_entry in self._functions.items() if function_entry.xcfg and not function_entry.minhash]
            return result
        if only_function_ids:
            result = [function_entry.function_id for _, function_entry in self._functions.items() if (not function_entry.minhash and function_entry.function_id in function_ids)]
        else:
            result = [function_entry for _, function_entry in self._functions.items() if (not function_entry.minhash and function_entry.function_id in function_ids)]
        return result

    def deleteXcfgData(self) -> None:
        for function_id, function_entry in self._functions.items():
            self._functions[function_id].xcfg = {}

    def deleteXcfgForSampleId(self, sample_id: int) -> None:
        for function_id, function_entry in self._functions.items():
            if sample_id == function_entry.sample_id:
                self._functions[function_id].xcfg = {}

    # TODO dict of what?
    def getContent(self) -> Dict[str, Any]:
        content = {
            "families": {family_id: family.toDict() for family_id, family in self._families.items()},
            "samples": {sample_id: sample.toDict() for sample_id, sample in self._samples.items()},
            "functions": {function_id: function.toDict() for function_id, function in self._functions.items()},
            "bands": self._bands,
        }
        return content

    # TODO dict of what?
    def setContent(self, content: Dict[str, Any]) -> None:
        self._families = {int(k): FamilyEntry.fromDict(v) for k, v in content["families"].items()}
        self._samples = {int(k): SampleEntry.fromDict(v) for k, v in content["samples"].items()}
        self._bands = {int(k): {int(ik): iv for ik, iv in v.items()} for k, v in content["bands"].items()}
        self._sample_by_sha256 = {sample.sha256: sample_id for sample_id, sample in self._samples.items()}
        self._sample_id_to_function_ids = defaultdict(list)
        self._pichashes = {}
        for function_id, function in content["functions"].items():
            function_id = int(function_id)
            function_entry = FunctionEntry.fromDict(function)
            self._functions[function_id] = function_entry
            self._sample_id_to_function_ids[function_entry.sample_id].append(function_id)
            if function_entry.pichash:
                if function_entry.pichash not in self._pichashes:
                    self._pichashes[function_entry.pichash] = set()
                self._pichashes[function_entry.pichash].add((function_entry.sample_id, function_entry.function_id))

    def getStats(self) -> Dict[str, Union[int, Dict[int, int]]]:
        stats = {
            "db_state": self._db_state,
            "num_families": len(self._families),
            "num_samples": len(self._samples),
            "num_functions": len(self._functions),
            "num_bands": len(self._bands),
            "num_pichashes": len(self._pichashes),
        }
        return stats

    def _addMinHashToBands(self, minhash: "MinHash") -> None:
        band_hashes = self.getBandHashesForMinHash(minhash)
        for band_number, band_hash in sorted(band_hashes.items()):
            if band_hash not in self._bands[band_number]:
                self._bands[band_number][band_hash] = []
            self._bands[band_number][band_hash].append(minhash.function_id)


    def getUniqueBlocks(self, sample_ids: Optional[List[int]] = None, progress_reporter=None) -> Dict:
        # query once to get all blocks from the functions of our samples
        block_statistics = {
            "by_sample_id": {
                sample_id: {
                    "sample_id": sample_id,
                    "total_blocks": 0,
                    "characteristic_blocks": 0,
                    "unique_blocks": 0
                } for sample_id in sample_ids
            },
            "unique_blocks_overall": 0,
            "num_samples": len(sample_ids)
        }
        candidate_picblockhashes = {}
        for function_id, entry in self._functions.items():
            sample_id = entry.sample_id
            for block_entry in entry.picblockhashes:
                block_hash = block_entry["hash"]
                if block_hash not in candidate_picblockhashes:
                    candidate_picblockhashes[block_hash] = {
                        "samples": set(),
                        "length": block_entry["length"],
                        "function_id": entry["function_id"],
                        "offset": block_entry["offset"],
                        "instructions": []
                    }
                candidate_picblockhashes[block_hash]["samples"].add(sample_id)
        # update statistics based on candidates
        for picblockhash, entry in candidate_picblockhashes.items():
            for sample_id in entry["samples"]:
                block_statistics["by_sample_id"][sample_id]["total_blocks"] += 1
        LOGGER.info(f"Found {len(candidate_picblockhashes)} candidate picblock hashes")
        for functiond_id, entry in self._functions.items():
            sample_id = entry.sample_id
            if sample_id not in sample_ids:
                for block_entry in entry.picblockhashes:
                    candidate_picblockhashes.pop(block_entry["hash"], None)
        # update statistics again after having reduced to results
        for picblockhash, entry in candidate_picblockhashes.items():
            if len(entry["samples"]) == 1:
                single_sample_id = list(entry["samples"])[0]
                block_statistics["by_sample_id"][single_sample_id]["unique_blocks"] += 1
            for sample_id in entry["samples"]:
                block_statistics["by_sample_id"][sample_id]["characteristic_blocks"] += 1
        block_statistics["unique_blocks_overall"] = len(candidate_picblockhashes)
        LOGGER.info(f"Reduced to {len(candidate_picblockhashes)} unique picblock hashes")
        # iterate over candidates by function_id and extract instructions
        function_id_to_block_offsets = {}
        for picblockhash, entry in candidate_picblockhashes.items():
            candidate_picblockhashes[picblockhash]["samples"] = sorted(list(entry["samples"]))
            if entry["function_id"] not in function_id_to_block_offsets:
                function_id_to_block_offsets[entry["function_id"]] = []
            function_id_to_block_offsets[entry["function_id"]].append((entry["offset"], picblockhash))
        for function_id, entry in self._functions:
            if function_id not in function_id_to_block_offsets.keys():
                continue
            for block_offset, picblockhash in function_id_to_block_offsets[function_id]:
                candidate_picblockhashes[picblockhash]["instructions"] = entry.xcfg["blocks"][str(block_offset)]
        return {"statistics": block_statistics, "unique_blocks": candidate_picblockhashes}

    def rebuildMinhashBandIndex(self, progress_reporter=None):
        # TODO while minhashes are considerably small, there is a still chance that the 
        # sum of all minhashes will eventually exceed available memory on a given system.
        # in this case we can start doing batches in our database interation already
        minhashes = []
        # get all minhash objects from all functions
        for function_id, function_entry in self._functions.items():
            if function_entry.minhash:
                minhash_obj = function_entry.getMinHash(self._minhash_config.MINHASH_SIGNATURE_BITS)
                minhashes.append(minhash_obj)
        # drop band collections
        # recreate indices for band collections
        collections = []
        for band_id in range(self._storage_config.STORAGE_NUM_BANDS):
            self._bands[band_id] = {}
        # re-add minhashes in batches
        batch_size = self._minhash_config.MINHASH_BAND_REBUILD_WORK_PACKAGE_SIZE
        if progress_reporter:
            progress_reporter.set_total((len(minhashes) // batch_size) + 1)
        for minhash_batch in zip_longest(*[iter(minhashes)]*batch_size):
            minhash_batch = [mh for mh in minhash_batch if mh is not None]
            for mh in minhash_batch:
                self._addMinHashToBands(mh)
            if progress_reporter:
                progress_reporter.step()
        return len(minhashes)

    ##### helpers for search ######

    @staticmethod
    def _get_sort_key_from_cursor(full_cursor: Optional[FullSearchCursor]):
        assert full_cursor is not None

        # Mapping entry -> tuple to sort by
        def get_tuple_from_entry(entry):
            return tuple(_get_field(entry, field) for field in full_cursor.sort_fields)

        # custom tuple compare function considering sort direction per dimension 
        is_backward_search = not full_cursor.is_forward_search
        sort_directions = [1 if direction ^ is_backward_search else -1 for direction in full_cursor.sort_directions]
        def compare_tuple(tuple1, tuple2):
            for i in range(len(full_cursor.sort_fields)):
                direction = sort_directions[i]
                val1 = tuple1[i]
                val2 = tuple2[i]
                if val1 == val2:
                    pass
                if val1 is None:
                    return -direction
                if val2 is None:
                    return +direction
                if val1 < val2:
                    return -direction
                if val1 > val2:
                    return direction
            return 0

        return lambda entry: functools.cmp_to_key(compare_tuple)(get_tuple_from_entry(entry))


    def _get_search_filter(self, search_fields:List[str], search_tree: NodeType, cursor: Optional[FullSearchCursor], conditional_search_fields=None) -> Callable:
        if cursor is not None:
            full_tree = AndNode([search_tree, cursor.toTree()])
        else:
            full_tree = search_tree
        full_tree = SearchFieldResolver(search_fields, conditional_search_fields=conditional_search_fields).visit(full_tree)
        full_tree = FilterSingleElementLists().visit(full_tree)
        full_tree = PropagateNot().visit(full_tree)
        filter = MemorySearchTranspiler().visit(full_tree)
        return filter

    ##### search ######

    # NOTE: Slowest part of search (by far) is the sorting. Maybe this can be precomputed or cached?
    # NOTE: Once sorting is fast, the search cursor could be handled separately from the search tree.
    #       Instead of putting it in the filter function, we could apply binary search on the sorted data.

    def findFamilyByString(self, search_tree: NodeType, cursor: Optional[FullSearchCursor] = None, max_num_results: int = 100) -> Dict[int, "FamilyEntry"]:
        result_dict = {}
        search_fields = ["family_name"]
        filter = self._get_search_filter(search_fields, search_tree, cursor)
        sort_key = self._get_sort_key_from_cursor(cursor)
        for entry in sorted(self._families.values(), key=sort_key):
            if filter(entry):
                result_dict[entry.family_id] = entry
            if len(result_dict) >= max_num_results:
                break
        return result_dict

    def findSampleByString(self, search_tree: NodeType, cursor: Optional[FullSearchCursor] = None, max_num_results: int = 100) -> Dict[int, "SampleEntry"]:
        result_dict = {}
        search_fields = ["filename", "family", "component", "version",]
        conditional_field = ("sha256", lambda search_term: len(search_term)>=3)
        filter = self._get_search_filter(search_fields, search_tree, cursor, conditional_search_fields=[conditional_field])
        sort_key = self._get_sort_key_from_cursor(cursor)
        for entry in sorted(self._samples.values(), key=sort_key):
            if filter(entry):
                result_dict[entry.sample_id] = entry
            if len(result_dict) >= max_num_results:
                break
        return result_dict

    def findFunctionByString(self, search_tree: NodeType, cursor: Optional[FullSearchCursor] = None, max_num_results: int = 100) -> Dict[int, "FunctionEntry"]:
        result_dict = {}
        # TODO also search through function labels once we have implemented them
        search_fields = ["function_name"]
        filter = self._get_search_filter(search_fields, search_tree, cursor)
        sort_key = self._get_sort_key_from_cursor(cursor)
        for entry in sorted(self._functions.values(), key=sort_key):
            if filter(entry):
                result_dict[entry.function_id] = entry
            if len(result_dict) >= max_num_results:
                break
        return result_dict


from typing import TYPE_CHECKING, Dict, List, Optional

from mcrit.storage.SampleEntry import SampleEntry
import mcrit.matchers.MatcherInterface as MatcherInterface

if TYPE_CHECKING:  # pragma: no cover
    from mcrit.storage.SampleEntry import SampleEntry

# Dataclass, post init
# constructor -> .fromSmdaFunction
# assume sample_entry, smda_function always available

class MatchedFunctionEntry(object):
    # basic information
    function_id: int
    num_bytes: int
    offset: int
    matched_family_id: int
    matched_family: str
    matched_sample_id: int
    matched_function_id: int
    matched_score: float
    matched_link_score: float
    matched_unique: bool
    matched_offset: int
    match_is_minhash: bool
    match_is_pichash: bool
    match_is_library: bool


    def __init__(self, function_id: int, num_bytes:int, offset:int, match_tuple: List) -> None:
        self.function_id = function_id
        self.num_bytes = num_bytes
        self.offset = offset
        self.matched_family_id = match_tuple[0]
        self.matched_sample_id = match_tuple[1]
        self.matched_function_id = match_tuple[2]
        self.matched_score = match_tuple[3]
        self.match_is_minhash = match_tuple[4] & MatcherInterface.IS_MINHASH_FLAG
        self.match_is_pichash = match_tuple[4] & MatcherInterface.IS_PICHASH_FLAG
        self.match_is_library = match_tuple[4] & MatcherInterface.IS_LIBRARY_FLAG
        self.matched_family = None
        self.matched_link_score = 0
        self.matched_unique = None
        self.matched_offset = None

    def getMatchTuple(self):
        return [
                self.matched_family_id,
                self.matched_sample_id,
                self.matched_function_id,
                self.matched_score,
                self.match_is_minhash * MatcherInterface.IS_MINHASH_FLAG
                + self.match_is_pichash * MatcherInterface.IS_PICHASH_FLAG
                + self.match_is_library * MatcherInterface.IS_LIBRARY_FLAG
            ]

    def toDict(self):
        matching_entry = {
            "fid": self.function_id,
            "num_bytes": self.num_bytes,
            "offset": self.offset,
            "matches": self.getMatchTuple()
        }
        return matching_entry

    @classmethod
    def fromDict(cls, entry_dict):
        matching_entry = cls(entry_dict["fid"], entry_dict["num_bytes"], entry_dict["offset"], entry_dict["matches"])
        return matching_entry

    def __str__(self):
        flag_str = "m" if self.match_is_minhash else "."
        flag_str += "p" if self.match_is_pichash else "."
        flag_str += "l" if self.match_is_library else "."
        return "Function: fid({}) num_bytes({}) - Matched: family_id({}) sample_id({}) function_id({}) score({}) flags({})".format(
            self.function_id,
            self.num_bytes,
            self.matched_family_id,
            self.matched_sample_id,
            self.matched_function_id,
            self.matched_score,
            flag_str
        )



import random
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple, Union

from mcrit.minhash.MinHash import MinHash

if TYPE_CHECKING: # pragma: no cover
    from mcrit.config.StorageConfig import StorageConfig
    from mcrit.config.McritConfig import McritConfig
    from mcrit.config.MinHashConfig import MinHashConfig
    from mcrit.storage.FunctionEntry import FunctionEntry
    from mcrit.storage.MatchingCache import MatchingCache
    from mcrit.storage.MemoryStorage import MemoryStorage
    from mcrit.storage.SampleEntry import SampleEntry
    from smda.common.SmdaFunction import SmdaFunction
    from smda.common.SmdaReport import SmdaReport

SampleId = int
FunctionId = int
FamilyId = int
PicHash = int
FamilyId = int
BandId = int
BandHash = int
PicHash = int
Sha256 = str


class StorageInterface:
    _config: "StorageConfig"
    _band_projection: None

    def __init__(self, config: "McritConfig") -> None:
        """Set the StorageConfig, sets up an empty Storage or loads existing data, ensures indexing,
        and ensures the existence of an Family with name \"\" and family_id 0.

        Args:
            config: the configuration of this Storage

        Raises:
            AssertionError: If family_id 0 already exists, but does not refer to Family \"\".
        """
        self._config = config
        self._storage_config = config.STORAGE_CONFIG
        self._minhash_config = config.MINHASH_CONFIG
        self._band_projection = None

    def dbLogEvent(self, event_msg: str, username: Optional[str] = None, details: Optional[dict] = None) -> None:
        """ Log an event to the database

        Args:
            event_msg: a string describing the event
            username: (optional) a username tied to this event
            details: a dict with arbitrary additional information
        """

    # -> Set[function_id]
    def getCandidatesForMinHash(self, minhash: "MinHash", band_matches_required=1) -> Set[int]:
        """Given a MinHash, return all candidates from all matching bands.

        Args:
            minhash: a MinHash
            band_matches_required: the number of bands a minhash needs to match before being considered a candidate

        Returns:
            candidates: a set of function_ids
        """
        raise NotImplementedError

    # -> Dict[function_id, Set[function_id]]
    def getCandidatesForMinHashes(self, function_id_to_minhash: Dict[int, "MinHash"], band_matches_required=1) -> Dict[int, Set[int]]:
        """Given MinHashes by function_id, return all candidates from all matching bands.

        Args:
            function_id_to_minhash: a dict mapping a function_id to a MinHash.
            band_matches_required: the number of bands a minhash needs to match before being considered a candidate

        Returns:
            candidates: a dict mapping a function_id to a set of candidate function_ids.
        """
        raise NotImplementedError

    def deleteXcfgData(self) -> None:
        """Delete XCFG data of all samples from the storage

        Returns:
            None
        """
        raise NotImplementedError

    def deleteXcfgForSampleId(self, sample_id: int) -> None:
        """Remove a sample's XCFG from the storage

        Args:
            sample_id: the sample_id whose XCFG will be deleted

        Returns:
            None
        """
        raise NotImplementedError

    def clearStorage(self) -> None:
        """Delete all contents in the storage, reinitialize an empty storage, ensure indexing and add Family \"\" with family_id.

        Returns:
            None
        """
        raise NotImplementedError

    def getSampleBySha256(self, sha256: str) -> Optional["SampleEntry"]:
        """Check if the given SHA256 is already associated with a sample in the storage

        Args:
            sha256: The SHA256 value to look up.

        Returns:
            A SampleEntry with the specified SHA256 or None, if no such SampleEntry exists.
        """
        raise NotImplementedError

    def addSmdaReport(self, smda_report: "SmdaReport", isQuery=False) -> Optional["SampleEntry"]:
        """Add a SMDA report to storage to create its sample representation and return corresponding SampleEntry object.
        This also adds the sample's functions. If the sample's family is not in storage yet, it will be added.
        If a sample with the same SHA256 already exists in storage, smda_report is not added and None is returned.

        Args:
            smda_report: the SmdaReport to add to the storage
            isQuery: True if this SmdaReport is associated with a matching query and its data shall not be persisted in the main DB

        Returns:
            A SampleEntry corresponding to smda_report or None, if SmdaReport was already added.
        """
        raise NotImplementedError

    def updateFunctionLabels(self, smda_report: "SmdaReport", username: str) -> Optional["SampleEntry"]:
        """Use a given SMDA report to update all non-dummy function labels for the SampleEntry matching its SHA256
        If no matching SampleEntry exists, no update happens.

        Args:
            smda_report: the SmdaReport to be used for updating function labels
            username: a username that allows tracking who submitted the labels

        Returns:
            True if SampleEntry was found and potentially labels were added, False otherwise.
        """
        raise NotImplementedError

    def importSampleEntry(self, sample_entry: "SampleEntry") -> Optional["SampleEntry"]:
        """Import a sample_entry to storage based on a previously exported SampleEntry.
        We assume that the family_id was already remapped by the MinHashIndex, meaning that only the sample_id needs to be adjusted
        If a sample with the same SHA256 already exists in storage, sample_entry is not added and None is returned.

        Args:
            sample_entry: the SampleEntry to add to the storage

        Returns:
            An adjusted SampleEntry if successful or None, if SampleEntry was already added.
        """
        raise NotImplementedError

    def importFunctionEntry(self, function_entry: "FunctionEntry") -> Optional["FunctionEntry"]:
        """Import a function_entry to storage based on a previously exported FunctionEntry.
        We assume that the family_id and sample_id were already remapped by the MinHashIndex, meaning that only the function_id needs to be adjusted.

        Args:
            function_entry: the FunctionEntry to add to the storage

        Returns:
            An adjusted FunctionEntry if successful or None, if we failed somewhere in the procedure
        """
        raise NotImplementedError


    def importFunctionEntries(self, function_entries: List["FunctionEntry"]) -> Optional[List["FunctionEntry"]]:
        """Import multiple function_entries to storage based on previously exported FunctionEntry objects.
        We assume that the family_id and sample_id were already remapped by the MinHashIndex, meaning that only the function_ids need to be adjusted.

        Args:
            function_entries: List of FunctionEntry objects to add to the storage

        Returns:
            A list of adjusted FunctionEntry objects if successful or None, if we failed somewhere in the procedure
        """
        raise NotImplementedError

    def modifyFamily(self, family_id: int, update_information: dict) -> bool:
        """Update a family from the storage

        Args:
            family_id: the id of the sample to modify
            update_information: a dictionary with update information for fields (family_name, is_library)

        Returns:
            True if family_id was contained in the storage and updated successfully, False otherwise
        """
        raise NotImplementedError

    def modifySample(self, sample_id: int, update_information: dict) -> bool:
        """Update a sample from the storage

        Args:
            sample_id: the id of the sample to modify
            update_information: a dictionary with update information for fields (family_name, version, component, is_library)

        Returns:
            True if sample_id was contained in the storage and updated successfully, False otherwise
        """
        raise NotImplementedError

    def deleteSample(self, sample_id: int) -> bool:
        """Remove a sample from the storage, also removes all functions of the sample.
        All minhashes will be removed from the bands.

        Args:
            sample_id: the id of the sample to delete

        Returns:
            True if sample_id was contained in the storage and deleted successfully, False otherwise
        """
        raise NotImplementedError

    def getSampleIds(self) -> List[int]:
        """Return a list of all sample_ids.

        Returns:
            a list of all sample_ids
        """
        raise NotImplementedError

    def getSampleIdByFunctionId(self, function_id: int) -> Optional[int]:
        """For a given function_id, return the corresponding sample_id or None, if function_id was not found.

        Args:
            function_id: a function_id

        Returns:
            the corresponding sample_id or None, if function_id was not found
        """
        raise NotImplementedError

    def getSampleById(self, sample_id: int) -> Optional["SampleEntry"]:
        """Given a sample_id, return the respective SampleEntry or None, if sample_id was not found.

        Args:
            sample_id: a sample_id

        Returns:
            the respective SampleEntry or None, if sample_id was not found
        """
        raise NotImplementedError

    def getSamplesByFamilyId(self, family_id: int) -> Optional[List["SampleEntry"]]:
        """Return all SampleEntries for a given family_id.

        Args:
            family_id: a family_id

        Returns:
            a list of all SampleEntries belonging to the given family_id.
            For a nonexistant family_id None is returned.
        """
        raise NotImplementedError

    def getSamples(self, start_index: int, limit: int) -> Optional["SampleEntry"]:
        """Iterate the sample collection and provide a slice (regardless of sample_id), 
        covering up to <limit> items, starting from start_index

        Args:
            start_index: <n>th sample in the collection
            limit: number of entries to return at most

        Returns:
            the respective SampleEntries

        """
        raise NotImplementedError

    def getLibraryInfoForSampleId(self, sample_id: int) -> Optional[Dict[str, str]]:
        """Return family_name and version of sample if it is a library or None otherwise

        Args:
            sample_id: a sample id

        Returns:
            If sample_id exists in storage and belongs to a library: a dict with keys
                "family" and "version" containing the respective data from the sample_id's SampeEnty, otherwise None.
        """
        raise NotImplementedError

    def getFunctionsBySampleId(self, sample_id: int) -> Optional[List["FunctionEntry"]]:
        """For a given sample_id, get all corresponding FunctionEntries.

        Args:
            sample_id: a sample_id

        Returns:
            A list of FunctionEntries or None, if sample_id does not exist
        """
        raise NotImplementedError

    def getFunctionIdsBySampleId(self, sample_id: int) -> Optional[List["int"]]:
        """For a given sample_id, get all corresponding function_ids.

        Args:
            sample_id: a sample_id

        Returns:
            A list of int or None, if sample_id does not exist
        """
        raise NotImplementedError

    def getFunctionById(self, function_id: int, with_xcfg=False) -> Optional["FunctionEntry"]:
        """Given a function_id, return the respective FunctionEntry or None, if function_id is not contained otherwise.

        Args:
            function_id: a function id
            with_xcfg: include xcfg info (default: False)

        Returns:
            the respective FunctionEntry or None, if function_id is not contained otherwise

        """
        raise NotImplementedError

    def getFunctions(self, start_index: int, limit: int) -> Optional["FunctionEntry"]:
        """Iterate the function collection and provide a slice (regardless of function_id), 
        covering up to <limit> items, starting from start_index

        Args:
            start_index: <n> function in the collection
            limit: number of entries to return at most

        Returns:
            the respective FunctionEntries

        """
        raise NotImplementedError

    # TODO remove this?
    def clearMatchingCache(self) -> None:
        """Clears the temporary matching cache

        Returns:
            None
        """
        raise NotImplementedError

    # TODO: make a MatchingCacheInterface, or MemoryStorage a subclass of MatchingCache?
    # TODO rename -> get?
    def createMatchingCache(self, function_ids: List[int]) -> Union["MemoryStorage", "MatchingCache"]:
        """Creates a temporary matching cache, for a list of function_ids

        Args:
            function_ids: list of function ids

        Returns:
            a matching cache for the specified list of function ids
        """
        raise NotImplementedError

    def isSampleId(self, sample_id: int) -> bool:
        """Check if a sample with sample_id exists in storage.

        Args:
            sample_id the sample id to check

        Returns:
            True if the sample_id exists, False otherwise
        """
        raise NotImplementedError

    def isFunctionId(self, function_id: int) -> bool:
        """Check if a function with function_id exists in storage.

        Args:
            function_id the function id to check

        Returns:
            True if the function_id exists, False otherwise
        """
        raise NotImplementedError

    def isFamilyId(self, family_id: int) -> bool:
        """Check if a family with family_id exists in storage.

        Args:
            family_id the family id to check

        Returns:
            True if the family_id exists, False otherwise
        """
        raise NotImplementedError

    def isPicHash(self, pichash: int) -> bool:
        """Check if a given PicHash exists in storage.

        Args:
            pichash: the pichash to check

        Returns:
            True if the PicHash exists, False otherwise
        """
        raise NotImplementedError

    # TODO check if a minhash can ever not have a function id
    def addMinHash(self, minhash: "MinHash") -> bool:
        """Add a MinHash object to the respective functions in storage.

        Args:
            minhash: the minhash to add

        Returns:
            True if the the respective function exists, and the MinHash could be added,
                False otherwise
        """
        raise NotImplementedError

    def addMinHashes(self, minhashes: List["MinHash"]) -> None:
        """Add multiple MinHash objects at once. MinHashes, that cannot be added
        (e.g. because of missing function ids) will be ignored.

        Args:
            minhashes: the list of MinHash objects to add

        Returns:
            None
        """
        raise NotImplementedError

    # TODO dict of what?
    def getContent(self, sample_ids: Optional[List[int]] = None) -> Dict[str, Any]:
        # filter sample_ids erlauben
        raise NotImplementedError

    # TODO dict of what?
    def addContent(self, content: Dict[str, Any]) -> None:
        # remaps ids, filters sha256
        raise NotImplementedError

    # TODO dict of what?
    def setContent(self, content: Dict[str, Any]) -> None:
        raise NotImplementedError

    def addFamily(self, family_name: str) -> int:
        """Add family, if not already known and return family_id

        Args:
            family_name: The name of the family to add

        Returns:
            family_id of the added (or already existing) family name
        """
        raise NotImplementedError

    def deleteFamily(self, family_id: int, keep_samples: Optional[str] = False) -> bool:
        """Delete family if known and return boolean success state

        Args:
            family_id: The family_id of the family to delete
            keep_samples: instead of deleting all samples and functions, reassign their family to "Unnamed (0)"

        Returns:
            True if successful
        """
        raise NotImplementedError

    def getFamilyIds(self) -> List[int]:
        """Returns a List of all family ids

        Returns:
            a List of all family ids
        """
        raise NotImplementedError

    def getFamilyId(self, family_name: str) -> Optional[int]:
        """Return family if known, None otherwise

        Args:
            family_name: Name of the family to look up

        Returns:
            family id corresponding to family_name or None
        """
        raise NotImplementedError

    def getFamily(self, family_id: int) -> Optional[str]:
        """Get the Family Name corresponding to a family id. Returns None, if family_id is not in storage.

        Args:
            family_id: the id of the family to look up

        Returns:
            the name of the family or None

        """
        raise NotImplementedError

    # TODO find out if it is really possible that a Function Object has no MinHash.
    def getMinHashByFunctionId(self, function_id: int) -> Optional[bytes]:
        """Get the MinHash's bytes of a function, if the function exists and has a MinHash object.

        Args:
            function_id: the id of the function

        Returns:
            The function's MinHash's bytes, or None

        """
        raise NotImplementedError

    # TODO can this happen, that the Function has no pichash?
    # TODO Maybe remove this function, or restore original behavior?
    # -> Dict[pichash, Set[Tuple[sample_id, function_id]]]
    def getPicHashMatchesByFunctionId(self, function_id: int) -> Optional[Dict[int, Set[Tuple[int, int]]]]:
        """Get the PicHash for a given function_id and also return all other sample_id/function_id pairs having the same PicHash

        Args:
            function_id: the function id for PicHash matching

        Returns:
            None if the function_id does not exist, or does not have a pichash. Otherwise a dict containing the functions pichash as key and the following value: a set of (sample_id, function_id) having the given pichash
        """
        raise NotImplementedError

    def getPicHashMatchesByFunctionIds(self, function_ids: List[int]) -> Dict[int, Set[Tuple[int, int]]]:
        """Get the PicHashes for given function_ids and also return all other sample_id/function_id pairs having the same PicHashes
        Nonexisting function ids or functions without pichashes will be ignored.

        Args:
            function_ids: the function ids for PicHash matching

        Returns:
            a dict mapping the function's pichashes to a set of (sample_id, function_id) having the same pichash.
        """
        raise NotImplementedError

    def getPicHashMatchesBySampleId(self, sample_id: int) -> Optional[Dict[int, Set[Tuple[int, int]]]]:
        """Get the PicHashes for a given sample_id's function_ids and also return all other sample_id/function_id pairs having the same PicHashes.
        Functions without pichashes will be ignored.
        If sample_id is not in storage, None will be returned

        Args:
            sample_id: the sample id for PicHash matching

        Returns:
            a dict mapping the sample's function's pichashes to a set of (sample_id, function_id) having the same pichash or None.
        """
        raise NotImplementedError

    def getMatchesForPicHash(self, pichash: int) -> Set[Tuple[int, int, int]]:
        """Get the set of all (family_id, sample_id, function_id) tuples for a given PicHash. If no function has the given pichash, the empty set is returned

        Args:
            pichash: the pichash to look up

        Returns:
            a set of (family_id, sample_id, function_id) tuples with the given pichash
        """
        raise NotImplementedError


    def getMatchesForPicBlockHash(self, picblockhash: int) -> Set[Tuple[int, int, int, int]]:
        """Get the set of all (family_id, sample_id, function_id, offset) tuples for a given PicBlockHash. If no function has the given picblockhash, the empty set is returned

        Args:
            picblockhash: the picblockhash to look up

        Returns:
            a set of (family_id, sample_id, function_id, offset) tuples with the given picblockhash
        """
        raise NotImplementedError

    def getStats(self) -> Dict[str, Union[int, Dict[int, int]]]:
        raise NotImplementedError

    def getUnhashedFunctions(self, function_ids: Optional[List[int]] = None, only_function_ids=False) -> List["FunctionEntry"]:
        """Given a list of function_ids, return all FunctionEntry objects corresponding to these IDs if they do not have a minhash yet.
        Otherwise, return all FunctionEntry objects that do not have a minhash.

        Args:
            function_ids: (optional) a list of function_ids
            only_function_ids: (optional) instead if FunctionEntry objects, only return function_ids

        Returns:
            a list of FunctionEntry objects without minhash
        """
        raise NotImplementedError


    def getUniqueBlocks(self, sample_ids: Optional[List[int]] = None, progress_reporter=None) -> Dict:
        """Given a list of sample_ids, return all basic blocks that are only found in any of these samples (and no other samples in the storage)

        Args:
            sample_ids: (optional) a list of sample_ids
            progress_reporter: (optional) might be passed by worker to inquiry progress of this DB only operation

        Returns:
            a dictionary with the isolated blocks
        """
        raise NotImplementedError

    def findFamilyByString(self, needle: str, max_num_results: int = 100) -> Dict[int, str]:
        """Given a needle, return all families that contain the term we are searching.

        Args:
            needle: a search string
            max_num_results: (optional) maximum number of results to return, default 100

        Returns:
            a dictionary of family_id->famile_name containing the needle
        """
        raise NotImplementedError

    def findSampleByString(self, needle: str, max_num_results: int = 100) -> Dict[int, "SampleEntry"]:
        """Given a needle, return all SampleEntry objects that contain the term we are searching.

        Args:
            needle: a search string
            max_num_results: (optional) maximum number of results to return, default 100

        Returns:
            a list of sample_id->SampleEntry containing the needle
        """
        raise NotImplementedError

    def findFunctionByString(self, needle: str, max_num_results: int = 100) -> Dict[int, "FunctionEntry"]:
        """Given a needle, return all FunctionEntry objects that contain the term we are searching.

        Args:
            needle: a search string
            max_num_results: (optional) maximum number of results to return, default 100

        Returns:
            a dict of function_id->FunctionEntry containing the needle
        """
        raise NotImplementedError

    def rebuildMinhashBandIndex(self) -> int:
        """Drop the current band index and rebuild it from scratch
        Returns:
            the number of minhashes indexed
        """
        raise NotImplementedError

    def createBandhashProjection(self, minhash):
        """Calculate a projection for index permutation based on a given minhash
        Args:
            minhash: the MinHash used as reference
        Returns:
            a dict containing signature indices used for bandhashing by band id
        """
        band_projection = {}
        random.seed(self._storage_config.STORAGE_BAND_SEED)
        band_index = 0
        for band_size, num_bands in self._storage_config.STORAGE_BANDS.items():
            for _ in range(num_bands):
                index_sequence = [index for index in range(len(minhash.getMinHashInt()))]
                random.shuffle(index_sequence)
                band_projection[band_index] = index_sequence[:band_size]
                band_index += 1
        return band_projection

    # -> Dict[BandIndex, BandHash]
    def getBandHashesForMinHash(self, minhash: "MinHash") -> Dict[int, int]:
        """Calculate band hashes for a given minhash, based on config parameters (STORAGE_BAND_SEED, STORAGE_BANDS)
        Args:
            minhash: the MinHash for which the BandHashes will be calculated
        Returns:
            a dict containing BandHashes by BandId
        """
        if self._band_projection is None:
            self._band_projection = self.createBandhashProjection(minhash)
        band_hashes = {}
        minhash_data = minhash.getMinHashInt()
        for band_index, permutation in self._band_projection.items():
            band_data = [minhash_data[i] for i in permutation]
            hashed_band_data = MinHash.hashData(band_data, 0)
            band_hashes[band_index] = hashed_band_data
        return band_hashes



from dataclasses import asdict, dataclass, field
from typing import Dict

from mcrit.config.ConfigInterface import ConfigInterface, default_field
from mcrit.storage.StorageFactory import StorageFactory


@dataclass
class StorageConfig(ConfigInterface):
    # storage configuration, use "memory" for local testing or "alchemy" when working with larger data
    # STORAGE_METHOD = StorageFactory.STORAGE_METHOD_MEMORY
    STORAGE_METHOD: ... = StorageFactory.STORAGE_METHOD_MONGODB
    # Use this as endpoint for our server
    STORAGE_SERVER: str = "127.0.0.1"
    # By default, MongoDbStorage's DB's name and MongoQueue's DB's name are both "mcrit"
    # Changing one DB name here or at runtime DOES NOT change the other name!
    STORAGE_MONGODB_DBNAME: str = "mcrit"
    STORAGE_PORT: ... = None
    # Once MinHashes have been calculated, discard disassembly from function entries
    STORAGE_DROP_DISASSEMBLY: bool = False
    # random seed to be used when deriving sequences used as bands
    STORAGE_BAND_SEED: int = 0xDEADBEEF
    # Banding supports:
    #  * MemoryStorage: arbitrary banding configuration, multiple lengths
    #  * MongoDbStorage: arbitrary banding configuration, multiple lengths
    # configuration for bands, dict with size:number as structure - we allow mixed sizes to increase scatter effect and randomness
    _default_storage_bands = {4: 20}
    STORAGE_BANDS: Dict[int, int] = default_field(_default_storage_bands)
    # use a hashmap to cache all banding data - very memory intensive, but great speedups.
    STORAGE_CACHE: bool = False

    @property
    def STORAGE_NUM_BANDS(self):
        num_bands = 0
        if self.STORAGE_BANDS:
            num_bands = sum([value for value in self.STORAGE_BANDS.values()])
        return num_bands



from typing import Dict, Set, Tuple
from mcrit.matchers.MatcherInterface import MatcherInterface, add_duration


class MatcherSample(MatcherInterface):
    def _additional_setup(self):
        self._sample_to_lib_info = {}
        self._sample_id_to_entry = {}

    @add_duration
    def getMatchesForSample(self, sample_id: int):
        self._function_entries = self._storage.getFunctionsBySampleId(sample_id)
        self._sample_id = sample_id
        sample_entry = self._storage.getSampleById(sample_id)
        self._sample_info = sample_entry.toDict()

        return self._getMatchesRoutine()

    def _getPicHashMatches(self) -> Dict[int, Set[Tuple[int, int, int]]]:
        return self._storage.getPicHashMatchesBySampleId(self._sample_id)



from typing import Dict, List
import numpy as np
from fastcluster import linkage
from scipy.spatial.distance import squareform


class MatcherCross(object):

    def _seriation(self, Z, N, cur_index):
        """
        input:
            - Z is a hierarchical tree (dendrogram)
            - N is the number of points given to the clustering process
            - cur_index is the position in the tree for the recursive traversal
        output:
            - order implied by the hierarchical tree Z

        seriation computes the order implied by a hierarchical tree (dendrogram)
        """
        if cur_index < N:
            return [cur_index]
        else:
            left = int(Z[cur_index - N, 0])
            right = int(Z[cur_index - N, 1])
            return self._seriation(Z, N, left) + self._seriation(Z, N, right)

    def _compute_serial_matrix(self, dist_mat, method="ward"):
        """
        input:
            - dist_mat is a distance matrix
            - method = ["ward","single","average","complete"]
        output:
            - seriated_dist is the input dist_mat,
            but with re-ordered rows and columns
            according to the seriation, i.e. the
            order implied by the hierarchical tree
            - res_order is the order implied by
            the hierarchical tree
            - res_linkage is the hierarchical tree (dendrogram)

        compute_serial_matrix transforms a distance matrix into
        a sorted distance matrix according to the order implied
        by the hierarchical tree (dendrogram)
        """
        N = len(dist_mat)
        flat_dist_mat = squareform(dist_mat)
        res_linkage = linkage(flat_dist_mat, method=method, preserve_input=True)
        res_order = self._seriation(res_linkage, N, N + N - 2)
        seriated_dist = np.zeros((N, N))
        a, b = np.triu_indices(N, k=1)
        seriated_dist[a, b] = dist_mat[[res_order[i] for i in a], [res_order[j] for j in b]]
        seriated_dist[b, a] = seriated_dist[a, b]

        return seriated_dist, res_order, res_linkage

    def _calculate_clustered_sequence(self, matching_percent):
        # convert to dist_mat
        matrix_as_lists = []
        for sample_id, entries in sorted(matching_percent.items()):
            row = []
            for other_sample_id, percent in sorted(entries.items()):
                # we need a symmetric distance, so we use whatever distance is smaller for our sample pair
                distance = min(100 - percent, 100 - matching_percent[other_sample_id][sample_id])
                row.append(distance)
            matrix_as_lists.append(row)
        dist_mat = np.array(matrix_as_lists)
        # calculate clustering
        sorted_matrix, sorted_sequence, linkage = self._compute_serial_matrix(dist_mat, method="complete")
        # map sequence indices to sample_ids
        sample_id_sequence = sorted(matching_percent.keys())
        mapped_sequence = []
        for sorted_id in sorted_sequence:
            mapped_sequence.append(sample_id_sequence[sorted_id])
        return mapped_sequence  

    def _produce_cross_crompare(self, matching_reports: List[Dict], method=None, custom_sequence=None, cluster_sequence=False):
        """ Generate a n:n cross comparision of all reports provided

        Args:
            matching_reports: matching report (as dict) of type 1vsN for each sample to cross compare
            method: which similarity percentage to use ("unweighted", "score_weighted", "frequency_weighted")
            custom_sequence: use this order of sample_ids
            cluster_sequence: use hierarchical clustering to automatically determine a sequence
        """
        if method is None:
            method = "unweighted"
        sample_id_to_matches = {}
        sample_sequence = []
        # load all reports

        for report in matching_reports:
            sample_id = str(report["info"]["sample"]["sample_id"])
            sample_sequence.append(sample_id)
            sample_id_to_matches[sample_id] = report["matches"]["samples"]

        # in case we have no custom reordering, use input orderering
        if custom_sequence is not None:
            sample_sequence = custom_sequence
            print("Using custom sample sequence:")
            print(sample_sequence)
        sample_id_set = set(sample_sequence)

        matching_matches = {}
        matching_percent = {}
        clustered_sequence = []

        matching_percent = {i: {j: 0 for j in sample_sequence} for i in sample_sequence}
        matching_matches = {i: {j: 0 for j in sample_sequence} for i in sample_sequence}
        for sample_id, matches in sorted(sample_id_to_matches.items()):
            for matched_sample in matches:
                other_sample_id = matched_sample["sample_id"]
                if str(other_sample_id) in sample_id_set:
                    matching_percent[str(sample_id)][str(other_sample_id)] = matched_sample["matched"]["percent"][method]
                    matching_matches[str(sample_id)][str(other_sample_id)] = matched_sample["matched"]["functions"]["combined"]
            # TODO: 2022-06-30 Don't process samples which are not part of the cross-compare
            if not str(sample_id) in matching_percent:
                matching_percent[str(sample_id)] = {}
            matching_percent[str(sample_id)][str(sample_id)] = 100
        if cluster_sequence:
            clustered_sequence =  self._calculate_clustered_sequence(matching_percent)
        return {
            "matching_matches": matching_matches,
            "matching_percent": matching_percent,
            "clustered_sequence": clustered_sequence,
        }


    def create_result(self, matching_reports: List[Dict]):
        MATCHING_METHODS = [
            "unweighted", 
            "score_weighted", 
            "frequency_weighted",
            "nonlib_unweighted", 
            "nonlib_score_weighted", 
            "nonlib_frequency_weighted"
        ]

        # potentially define a custom sequence (list of int), sample_id
         # aggregate and produce output
        result = {}
        for matching_method in MATCHING_METHODS:
            result[matching_method] = self._produce_cross_crompare(
                matching_reports,
                method=matching_method,
                cluster_sequence=True
            )
        return result 


from collections import defaultdict

class Graph:
    # straightforward implementation to enable us to do DFS to find clusters in undirected graphs
    # credit: Neelam Yadav
    # https://www.geeksforgeeks.org/depth-first-search-or-dfs-for-a-graph/
    def __init__(self):
        self.graph = defaultdict(list)

    def addNode(self, v):
        self.graph[v].append(v)

    def addEdge(self, u, v):
        self.graph[u].append(v)

    def DFSUtil(self, v, visited):
        visited.add(v)
        for neighbour in self.graph[v]:
            if neighbour not in visited:
                self.DFSUtil(neighbour, visited)

    def DFS(self, v):
        visited = set()
        self.DFSUtil(v, visited)
        return visited



from calendar import c
import json
import traceback
import uuid
import logging
from collections import defaultdict
from datetime import datetime, timedelta

# Only do basicConfig if no handlers have been configured
if len(logging._handlerList) == 0:
    logging.basicConfig(level=logging.INFO, format="%(asctime)-15s %(message)s")
LOGGER = logging.getLogger(__name__)


# copied from mongoqueue
class Job(object):
    """
    This class is also used to represent jobs to the outside
    """
    def __init__(self, data, queue):
        self._data = data
        self._queue = queue
        # All currently known remote jobs as implemented in the Worker, redundantly organized by various aspects
        self.method_types = {
            "matching": [
                "getMatchesForSample",
                "getMatchesForSampleVs",
                "combineMatchesToCross",
            ],
            "query": [
                "getMatchesForUnmappedBinary",
                "getMatchesForMappedBinary",
                "getMatchesForSmdaReport",
            ],
            "blocks": [
                "getUniqueBlocks",
            ],
            "minhashing": [
                "updateMinHashesForSample",
                "updateMinHashes",
                "rebuildIndex",
            ],
            "collection": [
                "addBinarySample",
                "deleteSample",
                "modifySample",
                "deleteFamily",
                "modifyFamily",
            ],
            "sample_id": [
                "getMatchesForSampleVs",
                "getMatchesForSample",
                "updateMinHashesForSample",
                "deleteSample",
                "modifySample",
            ],
            "family_id": [
                "deleteFamily",
                "modifyFamily",
            ],
            "all": [
                "getMatchesForSample",
                "getMatchesForSampleVs",
                "combineMatchesToCross",
                "getMatchesForUnmappedBinary",
                "getMatchesForMappedBinary",
                "getMatchesForSmdaReport",
                "getUniqueBlocks",
                "updateMinHashesForSample",
                "updateMinHashes",
                "addBinarySample",
                "modifySample",
                "deleteSample",
                "modifyFamily",
                "deleteFamily",
                "rebuildIndex",
            ]
        }

    def __str__(self) -> str:
        return f"ID: {self.job_id} - {self.parameters} | created: {self.created_at}, finished: {self.finished_at}, result: {self.result}, progress: {self.progress}"
    
    @property
    def family(self) -> str:
        if self.method in ["addBinarySample"]:
            descriptor = json.loads(self._data["payload"]["descriptor"])
            return descriptor[1]["2"]

    @property
    def sha256(self) -> str:
        if self.method in ["getMatchesForUnmappedBinary", "getMatchesForMappedBinary", "getMatchesForSmdaReport", "addBinarySample"]:
            descriptor = json.loads(self._data["payload"]["descriptor"])
            return descriptor[2]["0"]
        
    @property
    def filename(self) -> str:
        if self.method in ["addBinarySample"]:
            descriptor = json.loads(self._data["payload"]["descriptor"])
            return descriptor[1]["1"]
    
    @property
    def family_id(self):
        if self.method in self.method_types["family_id"]:
            return int(self.arguments[0])
        if self.method == "getUniqueBlocks":
            if len(self.arguments) > 1:
                return int(self.arguments[1])

    @property
    def has_family_id(self):
        if self.method in self.method_types["family_id"] or self.method == "getUniqueBlocks":
            return True
        return False
    
    @property
    def sample_ids(self):
        sample_ids = []
        if self.method == "combineMatchesToCross":
            sample_ids = [int(key) for key, value in self.arguments[0].items()]
        elif self.method == "getUniqueBlocks":
            sample_ids = [int(sid) for sid in self.arguments[0]]
        elif self.method in self.method_types["sample_id"]:
            sample_ids.append(int(self.arguments[0]))
            if self.method == "getMatchesForSampleVs":
                sample_ids.append(int(self.arguments[1]))
        elif self.sample_id is not None:
            sample_ids.append(self.sample_id)
        return set(list(sample_ids))

    @property
    def sample_id(self):
        if self.method in self.method_types["sample_id"]:
            return int(self.arguments[0])
        elif self.method == "updateMinHashesForSample":
            return int(self.arguments[0])
        elif self.method == "getUniqueBlocks":
            return int(self.arguments[0][0])
        elif self.method == "deleteSample":
            return int(self.arguments[0])
        elif self.method == "modifySample":
            return int(self.arguments[0])

    @property
    def other_sample_id(self):
        if self.method == "getMatchesForSampleVs":
            return int(self.arguments[1])

    def has_sample_id(self, sample_id:int):
        if self.method in self.method_types["sample_id"]:
            if self.method == "getMatchesForSampleVs":
                return int(self.arguments[0]) == sample_id or int(self.arguments[1]) == sample_id
            else:
                return int(self.arguments[0]) == sample_id

    @property
    def is_matching_job(self):
        if self.method in self.method_types["matching"]:
            return True
        return False
    
    @property
    def is_minhashing_job(self):
        if self.method in self.method_types["minhashing"]:
            return True
        return False

    @property
    def is_query_job(self):
        if self.method in self.method_types["query"]:
            return True
        return False
    
    @property
    def is_block_job(self):
        if self.method in self.method_types["blocks"]:
            return True
        return False

    @property
    def is_collection_job(self):
        if self.method in self.method_types["collection"]:
            return True
        return False

    @property
    def method(self):
        if "payload" in self._data and "method" in self._data["payload"]:
            return self._data["payload"]["method"]

    @property
    def payload(self):
        return self._data["payload"]

    @property
    def arguments(self):
        combined_values = []
        if "payload" in self._data and "params" in self._data["payload"] and "method" in self._data["payload"]:
            payload_params = json.loads(self._data["payload"]["params"])
            indexed_key_values = []
            named_key_values = []
            for k, v in payload_params.items():
                try:
                    int(k)
                    indexed_key_values.append(v)
                except:
                    named_key_values.append(v)
            combined_values = indexed_key_values + named_key_values
        return combined_values

    @property
    def parameters(self):
        method_str = ""
        if "payload" in self._data and "params" in self._data["payload"] and "method" in self._data["payload"]:
            method_str = self._data["payload"]["method"]
            method_str += "(" + ", ".join([str(v) for v in self.arguments]) + ")"
        return method_str

    @property
    def all_dependencies(self):
        return self._data["all_dependencies"]

    @property
    def job_id(self):
        if isinstance(self._data["_id"], dict):
            return str(self._data["_id"]["$oid"])
        return self._data["_id"]

    @property
    def number(self):
        return self._data["number"] if "number" in self._data else -1

    @property
    def priority(self):
        return self._data["priority"]

    @property
    def attempts_left(self):
        return self._data["attempts_left"]

    @property
    def is_failed(self):
        return self._data["attempts_left"] == 0

    @property
    def locked_by(self):
        return self._data["locked_by"]

    @property
    def locked_at(self):
        if isinstance(self._data["locked_at"], dict):
            return str(self._data["locked_at"]["$date"])
        return self._data["locked_at"]

    @property
    def last_error(self):
        return self._data["last_error"]

    @property
    def finished_at(self):
        if isinstance(self._data["finished_at"], dict):
            return str(self._data["finished_at"]["$date"])
        return self._data["finished_at"]

    @property
    def is_finished(self):
        return self._data["finished_at"] is not None

    @property
    def duration(self):
        if self.is_finished:
            FMT = '%Y-%m-%d-%H:%M:%S'
            finished_at = self.finished_at[:10] + "-" + self.finished_at[11:19]
            started_at = self.started_at[:10] + "-" + self.started_at[11:19]
            duration = datetime.strptime(finished_at, FMT)-datetime.strptime(started_at, FMT)
            return duration
        return None

    @property
    def created_at(self):
        if isinstance(self._data["created_at"], dict):
            return str(self._data["created_at"]["$date"])
        return self._data["created_at"]

    @property
    def started_at(self):
        if isinstance(self._data["started_at"], dict):
            return str(self._data["started_at"]["$date"])
        return self._data["started_at"]

    @property
    def is_terminated(self):
        return self._is_terminated()

    @property
    def progress(self):
        return self._data["progress"] if "progress" in self._data else 0

    # NOTE: This is a GridFS id, not the actual result
    @property
    def result(self):
        return self._data["result"]

    @result.setter
    def result(self, res):
        self._data["result"] = res

    ## job control

    def complete(self, result=None):
        """job has been completed."""
        if result:
            self._data["result"] = result
        self._data["finished_at"] = datetime.now()
        self._data["progress"] = 1
        return self

    def error(self, message=None):
        """note an error processing a job, and return it to the queue."""
        self._data["locked_by"] = None
        self._data["locked_at"] = None
        self._data["last_error"] = message
        self._data["attempts_left"] -= 1

    def progressor(self, count=0):
        self._data["progress"] = count

    def __enter__(self):
        return self._data

    def __exit__(self, type, value, tb):
        if (type, value, tb) == (None, None, None):
            self.complete()
        else:
            error = traceback.format_exc()
            self.error(error)

    # only works for jobs that report progress
    # used_cached flag is useless for this kind of queue
    def _is_terminated(self, use_cached=False):
        terminated = self._data["terminated"]
        if terminated is None:
            return False
        return terminated

    # only works for jobs that report progress
    def terminate(self):
        self._data["terminated"] = True


class LocalQueue(object):
    def __init__(self):
        self._setup_empty_queue()
        self._worker = None
        self._job_counter = 0
        self.clean_interval = 10 ** 9
        self.cache_time = 10 ** 9
        self.max_attempts = 1

    def _setup_empty_queue(self):
        self._jobs = defaultdict(lambda: None)
        self._files = defaultdict(lambda: None)
        self._files_meta = defaultdict(lambda: None)
        self._descriptor_to_job = defaultdict(lambda: None)
        self._hash_to_file = defaultdict(lambda: None)

    def registerWorker(self):
        # left empty, as we only have one worker in local mode
        pass

    def unregisterWorker(self):
        # left empty, as we only have one worker in local mode
        pass

    def getQueueStatistics(self):
        # we don't track stats for the volatile local mode
        return {}

    def set_worker(self, worker):
        self._worker = worker

    def get_job(self, job_id):
        data = self._jobs[job_id]
        return data and Job(data, self)

    def get_jobs(self, start_index: int, limit: int, method=None, state=None, filter=None, ascencing=False):
        # TODO implement all the filtering methods properly
        jobs = []
        for job_id, job_document in self._jobs.items():
            if method is not None and job_document["payload"]["method"] != method:
                continue
            jobs.append(Job(job_document, self))
        return jobs[start_index:start_index+limit]

    def get_cached_job_id(self, payload):
        return self._descriptor_to_job[payload["descriptor"]]

    def _file_to_grid(self, file, metadata=None):
        id = str(uuid.uuid4())
        try:
            file.seek(0)
            data = file.read()
        except AttributeError:
            if not isinstance(file, (str, bytes)):
                raise TypeError("Only strings, bytes or file-like objecs are supported")
            if isinstance(file, str):
                try:
                    data = data.encode(self.encoding)
                except AttributeError:
                    raise TypeError("no encoding was specified")
            else:
                data = file
        self._files[id] = data
        self._files_meta[id] = metadata
        try:
            self._hash_to_file[metadata["sha256"]] = id
        except:
            pass
        return id

    def _grid_to_file(self, grid, results_only=True):
        file = self._files[grid]
        if file is None:
            return None
        if results_only:
            metadata = self._files_meta[grid]
            if not "result" in metadata or not metadata["result"]:
                return b'"Access Not Allowed"'
        return file

    def get_file_by_hash_inc_lock(self, sha256):
        id = self._hash_to_file[sha256]
        self._inc_lock(id)
        return id

    def _get_file_by_hash(self, sha256):
        return self._hash_to_file[sha256]

    def _inc_lock(self, id):
        if id is None:
            return
        try:
            self._files_meta[id]["tmp_lock"] += 1
        except:
            pass

    def _dec_lock(self, id):
        if id is None:
            return
        try:
            self._files_meta[id]["tmp_lock"] -= 1
        except:
            pass

    def add_job_id_to_file(self, job_id, file_id):
        self._files_meta[file_id]["jobs"].append(job_id)
        self._dec_lock(file_id)

    def _dicts_to_grid(self, dicts, **kwargs):
        return self._file_to_grid(json.dumps(dicts).encode("ascii"), **kwargs)

    def _grid_to_dicts(self, grid, **kwargs):
        return json.loads(self._grid_to_file(grid, **kwargs).decode("ascii"))

    def _grid_to_meta(self, grid):
        return self._files_meta[grid]

    def _delete_grid(self, grid):
        meta = self._files_meta[grid]
        try:
            if self._hash_to_file[meta["sha256"]] == grid:
                del self._hash_to_file[meta["sha256"]]
        except:
            pass
        del self._files[grid]
        del self._files_meta[grid]

    def put(self, payload, await_jobs=[]):
        id = str(uuid.uuid4())
        job_data = defaultdict(lambda: None)
        job_data["_id"] = id
        job_data["number"] = self._job_counter
        self._job_counter += 1
        job_data["payload"] = payload
        job_data["unfinished_dependencies"] = await_jobs
        job_data["all_dependencies"] = await_jobs
        job_data["attempts_left"] = self.max_attempts
        job_data["created_at"] = datetime.now()
        self._jobs[id] = job_data
        self._descriptor_to_job[payload["descriptor"]] = id
        job_data["started_at"] = datetime.now()
        # NOTE: we can just ignore await jobs, because all jobs are
        #       executed in submission order
        self._worker._executeJob(Job(job_data, self))
        return id

    def clear(self):
        self.terminate_all_jobs()
        self._setup_empty_queue()

    def delete_job(self, id):
        if id not in self._jobs:
            return 0
        job = self._jobs[id]
        result = job["result"]
        file_params = json.loads(job["payload"]["file_params"])
        descriptor = job["payload"]["descriptor"]
        if self._descriptor_to_job[descriptor] == id:
            del self._descriptor_to_job[descriptor]
        del self._jobs[id]
        self._delete_grid(result)
        for f in file_params.values():
            meta = self._grid_to_meta(f)
            LOGGER.debug("Job meta: %s", meta)
            meta["jobs"].remove(id)
        return 1
    delete_history = []

    def delete_jobs(self, method=None, created_before=None, finished_before=None):
        delete_list = []
        deleted_count = 0
        for d in self._jobs.values():
            if d["payload"]["method"] is not None and d["payload"]["method"] != method:
                continue
            if d["created_at"] is not None and d["created_at"] >= finished_before:
                continue
            if d["finished_at"] is not None and d["finished_at"] >= finished_before:
                continue
            delete_list.append(d["_id"])
        for id in delete_list:
            deleted_count += self.delete_job(id)
        return deleted_count

    def clean(self):
        time_threshold = datetime.now() - timedelta(seconds=self.cache_time)
        delete_list = []
        for d in self._jobs.values():
            if (d["finished_at"] is not None) and d["finished_at"] < time_threshold:
                delete_list.append(d["_id"])

        self.delete_history.append(delete_list)
        LOGGER.debug("Perfomed clean, delete_history now has %d entries.", len(self.delete_history))

        for id in delete_list:
            self.delete_job(id)

        delete_list = []
        for id, meta in self._files_meta.items():
            if "result" in meta and meta["result"]:
                continue
            if len(meta["jobs"]) == 0 and meta["tmp_lock"] == 0:
                delete_list.append(id)
        for id in delete_list:
            self._delete_grid(id)

    def terminate_all_jobs(self):
        for job in self._jobs.values():
            if job["result"] is None:
                job["terminated"] = True

    def next(self):
        # this is a dummy, as we don't have an actual queue
        return None